package lc.syntax.tree;

import java.util.Objects;
import java.util.OptionalInt;
import java.util.stream.Stream;

public record Abstraction(String parameter, Term term) implements Term {

    public Term apply(final Term argument) {
        // From Wikipedia:
        // the beta-reduction of (\V.M) N is M[V := N].
        return this.term.substitute(this.parameter, argument);
    }

    @Override
    public Stream<String> free() {
        // a lambda abstraction binds its parameter name
        return this.term.free().filter(name -> !Objects.equals(name, this.parameter));
    }

    @Override
    public Term substitute(final String name, final Term term) {
        if (Objects.equals(this.parameter, name)) {
            // only substitute free names (`name` is bound here => stop substitution)
            return this;
        } else {
            return new Abstraction(this.parameter, this.term.substitute(name, term));
        }
    }

    @Override
    public String show(final boolean numbers) {
        final StringBuilder builder = new StringBuilder();

        final Runnable runnable = () -> {
            builder.append('\u03BB');
            builder.append(parameter);
            builder.append('.');
            builder.append(term.show(numbers));
        };

        if (numbers) {
            this.toNumber().ifPresentOrElse(
                builder::append,
                runnable
            );
        } else {
            runnable.run();
        }

        return builder.toString();
    }

    @Override
    public OptionalInt toNumber() {
        if (this.term instanceof final Abstraction x) {
            final Variable fName = new Variable(this.parameter);
            final Variable xName = new Variable(x.parameter());

            Term encoding = x.term();
            int n = 0;
            while (encoding instanceof final Application application) {
                if (application.left().equals(fName)) {
                    encoding = application.right();
                    n += 1;
                } else {
                    break;
                }
            }

            if (encoding.equals(xName)) {
                return OptionalInt.of(n);
            }
        }

        return OptionalInt.empty();
    }

}
